"""PyPaperless common types."""

import datetime
from dataclasses import dataclass, field
from enum import Enum, StrEnum
from typing import Any


# custom_fields
class CustomFieldType(Enum):
    """Represent a subtype of `CustomField`."""

    STRING = "string"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    FLOAT = "float"
    MONETARY = "monetary"
    DATE = "date"
    URL = "url"
    DOCUMENT_LINK = "documentlink"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, *_: object) -> "CustomFieldType":
        """Set default member on unknown value."""
        return CustomFieldType.UNKNOWN


# documents
@dataclass(kw_only=True)
class CustomFieldValueType:
    """Represent a subtype of `Document`."""

    field: int | None = None
    value: Any | None = None


# documents
@dataclass(kw_only=True)
class DocumentMetadataType:
    """Represent a subtype of `DocumentMeta`."""

    namespace: str | None = None
    prefix: str | None = None
    key: str | None = None
    value: str | None = None


# documents
@dataclass(kw_only=True)
class DocumentSearchHitType:
    """Represent a subtype of `Document`."""

    score: float | None = None
    highlights: str | None = None
    note_highlights: str | None = None
    rank: int | None = None


# mixins/models/data_fields, used for classifiers
class MatchingAlgorithmType(Enum):
    """Represent a subtype of `Correspondent`, `DocumentType`, `StoragePath` and `Tag`."""

    NONE = 0
    ANY = 1
    ALL = 2
    LITERAL = 3
    REGEX = 4
    FUZZY = 5
    AUTO = 6
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "MatchingAlgorithmType":
        """Set default member on unknown value."""
        return MatchingAlgorithmType.UNKNOWN


# mixins/models/securable
@dataclass(kw_only=True)
class PermissionSetType:
    """Represent a Paperless permission set."""

    users: list[int] = field(default_factory=list)
    groups: list[int] = field(default_factory=list)


# mixins/models/securable
@dataclass(kw_only=True)
class PermissionTableType:
    """Represent a Paperless permissions type."""

    view: PermissionSetType = field(default_factory=PermissionSetType)
    change: PermissionSetType = field(default_factory=PermissionSetType)


# documents
class RetrieveFileMode(StrEnum):
    """Represent a subtype of `DownloadedDocument`."""

    DOWNLOAD = "download"
    PREVIEW = "preview"
    THUMBNAIL = "thumb"


# saved_views
@dataclass(kw_only=True)
class SavedViewFilterRuleType:
    """Represent a subtype of `SavedView`."""

    rule_type: int | None = None
    value: str | None = None


# share_links
class ShareLinkFileVersionType(Enum):
    """Represent a subtype of `ShareLink`."""

    ARCHIVE = "archive"
    ORIGINAL = "original"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, *_: object) -> "ShareLinkFileVersionType":
        """Set default member on unknown value."""
        return ShareLinkFileVersionType.UNKNOWN


# status
class StatusType(Enum):
    """Represent a subtype of `Status`."""

    OK = "OK"
    ERROR = "ERROR"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def _missing_(cls: type, *_: object) -> "StatusType":
        """Set default member on unknown value."""
        return StatusType.UNKNOWN


# status
@dataclass(kw_only=True)
class StatusDatabaseMigrationStatusType:
    """Represent a subtype of `StatusDatabaseType`."""

    latest_migration: str | None = None
    unapplied_migrations: list[str] = field(default_factory=list)


# status
@dataclass(kw_only=True)
class StatusDatabaseType:
    """Represent a subtype of `Status`."""

    type: str | None = None
    url: str | None = None
    status: StatusType | None = None
    error: str | None = None
    migration_status: StatusDatabaseMigrationStatusType | None = None


# status
@dataclass(kw_only=True)
class StatusStorageType:
    """Represent a subtype of `Status`."""

    total: int | None = None
    available: int | None = None


# status
@dataclass(kw_only=True)
class StatusTasksType:
    """Represent a subtype of `Status`."""

    redis_url: str | None = None
    redis_status: StatusType | None = None
    redis_error: str | None = None
    celery_status: StatusType | None = None
    index_status: StatusType | None = None
    index_last_modified: datetime.datetime | None = None
    index_error: str | None = None
    classifier_status: StatusType | None = None
    classifier_last_trained: datetime.datetime | None = None
    classifier_error: str | None = None


# tasks
class TaskStatusType(Enum):
    """Represent a subtype of `Task`."""

    PENDING = "PENDING"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def _missing_(cls: type, *_: object) -> "TaskStatusType":
        """Set default member on unknown value."""
        return TaskStatusType.UNKNOWN


# workflows
class WorkflowActionType(Enum):
    """Represent a subtype of `Workflow`."""

    ASSIGNMENT = 1
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowActionType":
        """Set default member on unknown value."""
        return WorkflowActionType.UNKNOWN


# workflows
class WorkflowTriggerType(Enum):
    """Represent a subtype of `Workflow`."""

    CONSUMPTION = 1
    DOCUMENT_ADDED = 2
    DOCUMENT_UPDATED = 3
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowTriggerType":
        """Set default member on unknown value."""
        return WorkflowTriggerType.UNKNOWN


# workflows
class WorkflowTriggerSourceType(Enum):
    """Represent a subtype of `Workflow`."""

    CONSUME_FOLDER = 1
    API_UPLOAD = 2
    MAIL_FETCH = 3
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowTriggerSourceType":
        """Set default member on unknown value."""
        return WorkflowTriggerSourceType.UNKNOWN



"""Provide base classes."""

from dataclasses import Field, dataclass, fields
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar, final

from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.models.utils import dict_value_to_object

if TYPE_CHECKING:
    from pypaperless import Paperless


ResourceT = TypeVar("ResourceT", bound="PaperlessModel")


class PaperlessBase:
    """Superclass for all classes in PyPaperless."""

    _api_path = API_PATH["index"]

    def __init__(self, api: "Paperless") -> None:
        """Initialize a `PaperlessBase` instance."""
        self._api = api


class HelperProtocol(Protocol, Generic[ResourceT]):
    """Protocol for any `HelperBase` instances and its ancestors."""

    _api: "Paperless"
    _api_path: str
    _resource: PaperlessResource
    _resource_cls: type[ResourceT]


class HelperBase(PaperlessBase, Generic[ResourceT]):
    """Base class for all helpers in PyPaperless."""

    _resource: PaperlessResource
    _resource_public: bool = True

    def __init__(self, api: "Paperless") -> None:
        """Initialize a `HelperBase` instance."""
        super().__init__(api)

        if self._resource_public:
            self._api.local_resources.add(self._resource)

    @property
    def is_available(self) -> bool:
        """Return if the attached endpoint is available, or not."""
        return self._resource in self._api.remote_resources


@dataclass(init=False)
class PaperlessModelProtocol(Protocol):
    """Protocol for any `PaperlessBase` instances and its ancestors."""

    _api: "Paperless"
    _api_path: str
    _data: dict[str, Any]
    _fetched: bool
    _params: dict[str, Any]

    # fmt: off
    def _get_dataclass_fields(self) -> list[Field]: ...
    def _set_dataclass_fields(self) -> None: ...
    # fmt: on


@dataclass(init=False)
class PaperlessModel(PaperlessBase):
    """Base class for all models in PyPaperless."""

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `PaperlessModel` instance."""
        super().__init__(api)

        self._data = {}
        self._data.update(data)
        self._fetched = False
        self._params: dict[str, Any] = {}

    @final
    @classmethod
    def create_with_data(
        cls: type[ResourceT],
        api: "Paperless",
        data: dict[str, Any],
        *,
        fetched: bool = False,
    ) -> ResourceT:
        """Return a new instance of `cls` from `data`.

        Primarily used by class factories to create new model instances.

        Example: `document = Document.create_with_data(...)`
        """
        item = cls(api, data=data)

        item._fetched = fetched  # noqa: SLF001
        if fetched:
            item._set_dataclass_fields()  # noqa: SLF001
        return item

    @final
    def _get_dataclass_fields(self) -> list[Field]:
        """Get the dataclass fields."""
        return [
            field
            for field in fields(self)
            if (not field.name.startswith("_") or field.name == "__search_hit__")
        ]

    @final
    def _set_dataclass_fields(self) -> None:
        """Set the dataclass fields from `self._data`."""
        for field in self._get_dataclass_fields():
            value = dict_value_to_object(
                f"{self.__class__.__name__}.{field.name}",
                self._data.get(field.name),
                field.type,
                field.default,
                self._api,
            )
            setattr(self, field.name, value)

    async def load(self) -> None:
        """Get `model data` from DRF."""
        data = await self._api.request_json("get", self._api_path, params=self._params)

        self._data.update(data)
        self._set_dataclass_fields()
        self._fetched = True



"""SecurableMixin for PyPaperless helpers."""


class SecurableMixin:
    """Provide the `request_full_permissions` property for PyPaperless helpers."""

    _request_full_perms: bool = False

    @property
    def request_permissions(self) -> bool:
        """Return whether the helper requests items with the `permissions` table, or not.

        Documentation: https://docs.paperless-ngx.com/api/#permissions
        """
        return self._request_full_perms

    @request_permissions.setter
    def request_permissions(self, value: bool) -> None:
        """Set whether the helper requests items with the `permissions` table, or not.

        Documentation: https://docs.paperless-ngx.com/api/#permissions
        """
        self._request_full_perms = value



"""CreatableMixin for PyPaperless models."""

from typing import Any, cast

from pypaperless.exceptions import DraftFieldRequiredError
from pypaperless.models.base import PaperlessModelProtocol
from pypaperless.models.utils import object_to_dict_value


class CreatableMixin(PaperlessModelProtocol):
    """Provide the `save` method for PyPaperless models."""

    _create_required_fields: set[str]

    async def save(self) -> int | str | tuple[int, int]:
        """Create a new `resource item` in Paperless.

        Return the created item `id`, or a `task_id` in case of documents.

        Example:
        -------
        ```python
        draft = paperless.documents.draft(document=bytes(...))
        draft.title = "Add a title"

        # request Paperless to store the new item
        draft.save()
        ```

        """
        self.validate()
        kwdict = self._serialize()
        res = await self._api.request_json("post", self._api_path, **kwdict)

        if type(self).__name__ == "DocumentNoteDraft":
            return (
                cast(int, max(item.get("id") for item in res)),
                cast(int, kwdict["json"]["document"]),
            )
        if isinstance(res, dict):
            return int(res["id"])
        return str(res)

    def _serialize(self) -> dict[str, Any]:
        """Serialize."""
        data = {
            "json": {
                field.name: object_to_dict_value(getattr(self, field.name))
                for field in self._get_dataclass_fields()
            },
        }
        # check for empty permissions as they will raise if None
        if "set_permissions" in data["json"] and data["json"]["set_permissions"] is None:
            del data["json"]["set_permissions"]

        return data

    def validate(self) -> None:
        """Check required fields before persisting the item to Paperless."""
        missing = [field for field in self._create_required_fields if getattr(self, field) is None]

        if len(missing) == 0:
            return

        message = f"Missing fields for saving a `{type(self).__name__}`: {', '.join(missing)}."
        raise DraftFieldRequiredError(message)



"""Provide `Correspondent`, `DocumentType`, `StoragePath` and `Tag` related models and helpers."""

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class Correspondent(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `Correspondent`."""

    _api_path = API_PATH["correspondents_single"]

    id: int | None = None
    slug: str | None = None
    name: str | None = None
    document_count: int | None = None
    last_correspondence: datetime.datetime | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Correspondent` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class CorrespondentDraft(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableDraftMixin,
    models.CreatableMixin,
):
    """Represent a new `Correspondent`, which is not yet stored in Paperless."""

    _api_path = API_PATH["correspondents"]

    _create_required_fields = {
        "name",
        "match",
        "matching_algorithm",
        "is_insensitive",
    }

    name: str | None = None


@dataclass(init=False)
class DocumentType(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `DocumentType`."""

    _api_path = API_PATH["document_types_single"]

    id: int | None = None
    slug: str | None = None
    name: str | None = None
    document_count: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentType` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class DocumentTypeDraft(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableDraftMixin,
    models.CreatableMixin,
):
    """Represent a new `DocumentType`, which is not yet stored in Paperless."""

    _api_path = API_PATH["document_types"]

    _create_required_fields = {
        "name",
        "match",
        "matching_algorithm",
        "is_insensitive",
    }

    name: str | None = None
    owner: int | None = None


@dataclass(init=False)
class StoragePath(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `StoragePath`."""

    _api_path = API_PATH["storage_paths_single"]

    id: int | None = None
    slug: str | None = None
    name: str | None = None
    path: str | None = None
    document_count: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `StoragePath` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class StoragePathDraft(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableDraftMixin,
    models.CreatableMixin,
):
    """Represent a new `StoragePath`, which is not yet stored in Paperless."""

    _api_path = API_PATH["storage_paths"]

    _create_required_fields = {
        "name",
        "path",
        "match",
        "matching_algorithm",
        "is_insensitive",
    }

    name: str | None = None
    path: str | None = None
    owner: int | None = None


@dataclass(init=False)
class Tag(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `Tag`."""

    _api_path = API_PATH["tags_single"]

    id: int | None = None
    slug: str | None = None
    name: str | None = None
    color: str | None = None
    text_color: str | None = None
    is_inbox_tag: bool | None = None
    document_count: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Tag` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class TagDraft(
    PaperlessModel,
    models.MatchingFieldsMixin,
    models.SecurableDraftMixin,
    models.CreatableMixin,
):
    """Represent a new `Tag`, which is not yet stored in Paperless."""

    _api_path = API_PATH["tags"]

    _create_required_fields = {
        "name",
        "color",
        "is_inbox_tag",
        "match",
        "matching_algorithm",
        "is_insensitive",
    }

    name: str | None = None
    color: str | None = None
    text_color: str | None = None
    is_inbox_tag: bool | None = None
    owner: int | None = None


class CorrespondentHelper(
    HelperBase[Correspondent],
    helpers.SecurableMixin,
    helpers.CallableMixin[Correspondent],
    helpers.DraftableMixin[CorrespondentDraft],
    helpers.IterableMixin[Correspondent],
):
    """Represent a factory for Paperless `Correspondent` models."""

    _api_path = API_PATH["correspondents"]
    _resource = PaperlessResource.CORRESPONDENTS

    _draft_cls = CorrespondentDraft
    _resource_cls = Correspondent


class DocumentTypeHelper(
    HelperBase[DocumentType],
    helpers.SecurableMixin,
    helpers.CallableMixin[DocumentType],
    helpers.DraftableMixin[DocumentTypeDraft],
    helpers.IterableMixin[DocumentType],
):
    """Represent a factory for Paperless `DocumentType` models."""

    _api_path = API_PATH["document_types"]
    _resource = PaperlessResource.DOCUMENT_TYPES

    _draft_cls = DocumentTypeDraft
    _resource_cls = DocumentType


class StoragePathHelper(
    HelperBase[StoragePath],
    helpers.SecurableMixin,
    helpers.CallableMixin[StoragePath],
    helpers.DraftableMixin[StoragePathDraft],
    helpers.IterableMixin[StoragePath],
):
    """Represent a factory for Paperless `StoragePath` models."""

    _api_path = API_PATH["storage_paths"]
    _resource = PaperlessResource.STORAGE_PATHS

    _draft_cls = StoragePathDraft
    _resource_cls = StoragePath


class TagHelper(
    HelperBase[Tag],
    helpers.SecurableMixin,
    helpers.CallableMixin[Tag],
    helpers.DraftableMixin[TagDraft],
    helpers.IterableMixin[Tag],
):
    """Represent a factory for Paperless `Tag` models."""

    _api_path = API_PATH["tags"]
    _resource = PaperlessResource.TAGS

    _draft_cls = TagDraft
    _resource_cls = Tag



"""Provide the PageGenerator class."""

from collections.abc import AsyncIterator
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Self

from pypaperless.models.base import PaperlessBase
from pypaperless.models.pages import Page

if TYPE_CHECKING:
    from pypaperless import Paperless


class PageGenerator(PaperlessBase, AsyncIterator):
    """Iterator for DRF paginated endpoints.

    `api`: An instance of :class:`Paperless`.
    `url`: A url returning DRF page contents.
    `resource`: A target resource model type for mapping results with.
    `params`: Optional dict of query string parameters.
    """

    _page: Page | None

    def __aiter__(self) -> Self:
        """Return self as iterator."""
        return self

    async def __anext__(self) -> Page:
        """Return next item from the current batch."""
        if self._page is not None and self._page.is_last_page:
            raise StopAsyncIteration

        res = await self._api.request_json("get", self._url, params=self.params)
        data = {
            **res,
            "_api_path": self._url,
            "current_page": self.params["page"],
            "page_size": self.params["page_size"],
        }
        self._page = Page.create_with_data(self._api, data, fetched=True)
        # dirty attach the resource to the data class
        self._page._resource_cls = self._resource_cls  # noqa: SLF001

        # rise page by one to request next page on next iteration
        self.params["page"] += 1

        # we do not reach this point without a self._page object, so: ignore type error
        return self._page

    def __init__(
        self,
        api: "Paperless",
        url: str,
        resource_cls: type,
        params: dict[str, Any] | None = None,
    ) -> None:
        """Initialize `PageGenerator` class instance."""
        super().__init__(api)

        self._page = None
        self._resource_cls = resource_cls
        self._url = url

        self.params = deepcopy(params) if params else {}
        self.params.setdefault("page", 1)
        self.params.setdefault("page_size", 150)



"""PyPaperless."""

import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from io import BytesIO
from json.decoder import JSONDecodeError
from typing import Any

import aiohttp
from yarl import URL

from . import helpers
from .const import API_PATH, PaperlessResource
from .exceptions import BadJsonResponseError, InitializationError, JsonResponseWithError
from .models.base import HelperBase


class Paperless:
    """Retrieves and manipulates data from and to Paperless via REST."""

    _helpers_map: set[tuple[str, type[HelperBase]]] = {
        (PaperlessResource.CONFIG, helpers.ConfigHelper),
        (PaperlessResource.CORRESPONDENTS, helpers.CorrespondentHelper),
        (PaperlessResource.CUSTOM_FIELDS, helpers.CustomFieldHelper),
        (PaperlessResource.DOCUMENTS, helpers.DocumentHelper),
        (PaperlessResource.DOCUMENT_TYPES, helpers.DocumentTypeHelper),
        (PaperlessResource.GROUPS, helpers.GroupHelper),
        (PaperlessResource.MAIL_ACCOUNTS, helpers.MailAccountHelper),
        (PaperlessResource.MAIL_RULES, helpers.MailRuleHelper),
        (PaperlessResource.SAVED_VIEWS, helpers.SavedViewHelper),
        (PaperlessResource.SHARE_LINKS, helpers.ShareLinkHelper),
        (PaperlessResource.STATUS, helpers.StatusHelper),
        (PaperlessResource.STORAGE_PATHS, helpers.StoragePathHelper),
        (PaperlessResource.TAGS, helpers.TagHelper),
        (PaperlessResource.TASKS, helpers.TaskHelper),
        (PaperlessResource.USERS, helpers.UserHelper),
        (PaperlessResource.WORKFLOWS, helpers.WorkflowHelper),
    }

    config: helpers.ConfigHelper
    correspondents: helpers.CorrespondentHelper
    custom_fields: helpers.CustomFieldHelper
    documents: helpers.DocumentHelper
    document_types: helpers.DocumentTypeHelper
    groups: helpers.GroupHelper
    mail_accounts: helpers.MailAccountHelper
    mail_rules: helpers.MailRuleHelper
    saved_views: helpers.SavedViewHelper
    share_links: helpers.ShareLinkHelper
    status: helpers.StatusHelper
    storage_paths: helpers.StoragePathHelper
    tags: helpers.TagHelper
    tasks: helpers.TaskHelper
    users: helpers.UserHelper
    workflows: helpers.WorkflowHelper

    async def __aenter__(self) -> "Paperless":
        """Return context manager."""
        await self.initialize()
        return self

    async def __aexit__(self, *_: object) -> None:
        """Exit context manager."""
        await self.close()

    def __init__(
        self,
        url: str | URL,
        token: str,
        *,
        session: aiohttp.ClientSession | None = None,
        request_args: dict[str, Any] | None = None,
    ) -> None:
        """Initialize a `Paperless` instance.

        You have to permit either a session, or an url / token pair.

        `url`: A hostname or IP-address as string, or yarl.URL object.
        `token`: An api token created in Paperless Django settings, or via the helper function.
        `session`: A custom `PaperlessSession` object, if existing.

        `request_args` are passed to each request method call as additional kwargs,
        ssl stuff for example. You should read the aiohttp docs to learn more about it.
        """
        self._base_url = self._create_base_url(url)
        self._initialized = False
        self._local_resources: set[PaperlessResource] = set()
        self._remote_resources: set[PaperlessResource] = set()
        self._request_args = request_args or {}
        self._session = session
        self._token = token
        self._version: str | None = None

        self.logger = logging.getLogger(f"{__package__}[{self._base_url.host}]")

    @property
    def is_initialized(self) -> bool:
        """Return `True` if connection is initialized."""
        return self._initialized

    @property
    def host_version(self) -> str | None:
        """Return the version object of the Paperless host."""
        return self._version

    @property
    def local_resources(self) -> set[PaperlessResource]:
        """Return a set of locally available resources."""
        return self._local_resources

    @property
    def remote_resources(self) -> set[PaperlessResource]:
        """Return a set of available resources of the Paperless host."""
        return self._remote_resources

    @staticmethod
    def _create_base_url(url: str | URL) -> URL:
        """Create URL from string or URL and prepare for further use."""
        # reverse compatibility, fall back to https
        if isinstance(url, str) and "://" not in url:
            url = f"https://{url}".rstrip("/")
        url = URL(url)

        # scheme check. fall back to https
        if url.scheme not in ("https", "http"):
            url = URL(url).with_scheme("https")

        return url

    @staticmethod
    def _process_form(data: dict[str, Any]) -> aiohttp.FormData:
        """Process form data and create a `aiohttp.FormData` object.

        Every field item gets converted to a string-like object.
        """
        form = aiohttp.FormData(quote_fields=False)

        def _add_form_value(name: str | None, value: Any) -> Any:
            if value is None:
                return
            params = {}
            if isinstance(value, dict):
                for dict_key, dict_value in value.items():
                    _add_form_value(dict_key, dict_value)
                return
            if isinstance(value, list | set):
                for list_value in value:
                    _add_form_value(name, list_value)
                return
            if isinstance(value, tuple):
                if len(value) == 2:
                    params["filename"] = f"{value[1]}"
                value = value[0]
            if name is not None:
                form.add_field(
                    name, BytesIO(value) if isinstance(value, bytes) else f"{value}", **params
                )

        _add_form_value(None, data)
        return form

    @staticmethod
    async def generate_api_token(
        url: str,
        username: str,
        password: str,
        session: aiohttp.ClientSession | None = None,
    ) -> str:
        """Request Paperless to generate an api token for the given credentials.

        Warning: the request is plain and insecure. Don't use this in production
        environments or businesses.

        Warning: error handling is low for this method, as it is just a helper.

        Example:
        -------
        ```python
        token = Paperless.generate_api_token("example.com:8000", "api_user", "secret_password")

        paperless = Paperless("example.com:8000", token)
        # do something
        ```

        """
        external_session = session is not None
        session = session or aiohttp.ClientSession()
        try:
            url = url.rstrip("/")
            json = {
                "username": username,
                "password": password,
            }
            res = await session.request("post", f"{url}{API_PATH['token']}", json=json)
            data = await res.json()
            res.raise_for_status()
            return str(data["token"])
        except (JSONDecodeError, KeyError) as exc:
            message = "Token is missing in response."
            raise BadJsonResponseError(message) from exc
        except aiohttp.ClientResponseError as exc:
            raise JsonResponseWithError(payload={"error": data}) from exc
        finally:
            if not external_session:
                await session.close()

    async def close(self) -> None:
        """Clean up connection."""
        if self._session:
            await self._session.close()
        self.logger.info("Closed.")

    async def initialize(self) -> None:
        """Initialize the connection to DRF and fetch the endpoints."""
        async with self.request("get", API_PATH["index"]) as res:
            try:
                res.raise_for_status()
                payload = await res.json()
            except (aiohttp.ClientResponseError, ValueError) as exc:
                raise InitializationError from exc

            self._version = res.headers.get("x-version", None)
            self._remote_resources = set(map(PaperlessResource, payload))

        for attribute, helper in self._helpers_map:
            setattr(self, f"{attribute}", helper(self))

        unused = self._remote_resources.difference(self._local_resources)
        missing = self._local_resources.difference(self._remote_resources)

        if len(unused) > 0:
            self.logger.debug("Unused features: %s", ", ".join(unused))

        if len(missing) > 0:
            self.logger.warning("Outdated version detected.")
            self.logger.warning("Missing features: %s", ", ".join(missing))
            self.logger.warning("Consider pulling the latest version of Paperless-ngx.")

        self.logger.info("Initialized.")

        self._initialized = True

    @asynccontextmanager
    async def request(  # noqa: PLR0913
        self,
        method: str,
        path: str,
        json: dict[str, Any] | None = None,
        data: dict[str, Any] | aiohttp.FormData | None = None,
        form: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> AsyncGenerator[aiohttp.ClientResponse, None]:
        """Send a request to the Paperless api and return the `aiohttp.ClientResponse`.

        This method provides a little interface for utilizing `aiohttp.FormData`.

        `method`: A http method: get, post, patch, put, delete, head, options
        `path`: A path to the endpoint or a string url.
        `json`: A dict containing the json data.
        `data`: A dict containing the data to send in the request body.
        `form`: A dict with form data, which gets converted to `aiohttp.FormData`
        and replaces `data`.
        `params`: A dict with query parameters.
        `kwargs`: Optional attributes for the `aiohttp.ClientSession.request` method.
        """
        if self._session is None:
            self._session = aiohttp.ClientSession()

        # add headers
        self._session.headers.update(
            {
                "Accept": "application/json; version=2",
                "Authorization": f"Token {self._token}",
            }
        )

        # add request args
        kwargs.update(self._request_args)

        # overwrite data with a form, when there is a form payload
        if isinstance(form, dict):
            data = self._process_form(form)

        # add base path
        url = f"{self._base_url}{path}" if not path.startswith("http") else path

        res = await self._session.request(
            method=method,
            url=url,
            json=json,
            data=data,
            params=params,
            **kwargs,
        )
        self.logger.debug("%s (%d): %s", method.upper(), res.status, res.url)
        yield res

    async def request_json(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> Any:
        """Make a request to the api and parse response json to dict."""
        async with self.request(method, endpoint, **kwargs) as res:
            if res.content_type != "application/json":
                raise BadJsonResponseError(res)

            try:
                payload = await res.json()
            except ValueError:
                raise BadJsonResponseError(res) from None

            if res.status == 400:
                raise JsonResponseWithError(payload)

            res.raise_for_status()

        return payload



"""Provide `MailRule` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class MailAccount(
    PaperlessModel,
    models.SecurableMixin,
):
    """Represent a Paperless `MailAccount`."""

    _api_path = API_PATH["mail_accounts_single"]

    id: int | None = None
    name: str | None = None
    imap_server: str | None = None
    imap_port: int | None = None
    imap_security: int | None = None
    username: str | None = None
    # exclude that from the dataclass
    # password: str | None = None  # noqa: ERA001
    character_set: str | None = None
    is_token: bool | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `MailAccount` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class MailRule(
    PaperlessModel,
    models.SecurableMixin,
):
    """Represent a Paperless `MailRule`."""

    _api_path = API_PATH["mail_rules_single"]

    id: int | None = None
    name: str | None = None
    account: int | None = None
    folder: str | None = None
    filter_from: str | None = None
    filter_to: str | None = None
    filter_subject: str | None = None
    filter_body: str | None = None
    filter_attachment_filename_include: str | None = None
    filter_attachment_filename_exclude: str | None = None
    maximum_age: int | None = None
    action: int | None = None
    action_parameter: str | None = None
    assign_title_from: int | None = None
    assign_tags: list[int] | None = None
    assign_correspondent_from: int | None = None
    assign_correspondent: int | None = None
    assign_document_type: int | None = None
    assign_owner_from_rule: bool | None = None
    order: int | None = None
    attachment_type: int | None = None
    consumption_scope: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `MailRule` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class MailAccountHelper(
    HelperBase[MailAccount],
    helpers.CallableMixin[MailAccount],
    helpers.IterableMixin[MailAccount],
    helpers.SecurableMixin,
):
    """Represent a factory for Paperless `MailAccount` models."""

    _api_path = API_PATH["mail_accounts"]
    _resource = PaperlessResource.MAIL_ACCOUNTS

    _resource_cls = MailAccount


class MailRuleHelper(
    HelperBase[MailRule],
    helpers.CallableMixin[MailRule],
    helpers.IterableMixin[MailRule],
    helpers.SecurableMixin,
):
    """Represent a factory for Paperless `MailRule` models."""

    _api_path = API_PATH["mail_rules"]
    _resource = PaperlessResource.MAIL_RULES

    _resource_cls = MailRule



"""PermissionFieldsMixin for PyPaperless models."""

from dataclasses import dataclass

from pypaperless.models.common import MatchingAlgorithmType


@dataclass
class MatchingFieldsMixin:
    """Provide shared matching fields for PyPaperless models."""

    match: str | None = None
    matching_algorithm: MatchingAlgorithmType | None = None
    is_insensitive: bool | None = None



"""Mixins for PyPaperless helpers."""

from .callable import CallableMixin
from .draftable import DraftableMixin
from .iterable import IterableMixin
from .securable import SecurableMixin

__all__ = (
    "CallableMixin",
    "DraftableMixin",
    "IterableMixin",
    "SecurableMixin",
)



"""Provide `ShareLink` related models and helpers."""

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import ShareLinkFileVersionType
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class ShareLink(
    PaperlessModel,
    models.DeletableMixin,
    models.UpdatableMixin,
):
    """Represent a Paperless `ShareLink`."""

    _api_path = API_PATH["share_links_single"]

    id: int
    created: datetime.datetime | None = None
    expiration: datetime.datetime | None = None
    slug: str | None = None
    document: int | None = None
    file_version: ShareLinkFileVersionType | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `MailAccount` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class ShareLinkDraft(
    PaperlessModel,
    models.CreatableMixin,
):
    """Represent a new Paperless `ShareLink`, which is not stored in Paperless."""

    _api_path = API_PATH["share_links"]

    _create_required_fields = {"document", "file_version"}

    expiration: datetime.datetime | None = None
    document: int | None = None
    file_version: ShareLinkFileVersionType | None = None


class ShareLinkHelper(
    HelperBase[ShareLink],
    helpers.CallableMixin[ShareLink],
    helpers.DraftableMixin[ShareLinkDraft],
    helpers.IterableMixin[ShareLink],
):
    """Represent a factory for Paperless `ShareLink` models."""

    _api_path = API_PATH["share_links"]
    _resource = PaperlessResource.SHARE_LINKS

    _draft_cls = ShareLinkDraft
    _resource_cls = ShareLink



"""DeletableMixin for PyPaperless models."""

from pypaperless.models.base import PaperlessModelProtocol


class DeletableMixin(PaperlessModelProtocol):
    """Provide the `delete` method for PyPaperless models."""

    async def delete(self) -> bool:
        """Delete a `resource item` from DRF. There is no point of return.

        Return `True` when deletion was successful, `False` otherwise.

        Example:
        -------
        ```python
        # request a document
        document = await paperless.documents(42)

        if await document.delete():
            print("Successfully deleted the document!")
        ```

        """
        async with self._api.request("delete", self._api_path) as res:
            return res.status == 204



"""PyPaperless exceptions."""

from typing import Any


class PaperlessError(Exception):
    """Base exception for PyPaperless."""


# Sessions and requests


class InitializationError(PaperlessError):
    """Raise when initializing a `Paperless` instance without valid url or token."""


class BadJsonResponseError(PaperlessError):
    """Raise when response is no valid json."""


class JsonResponseWithError(PaperlessError):
    """Raise when Paperless accepted the request, but responded with an error payload."""

    def __init__(self, payload: Any) -> None:
        """Initialize a `JsonResponseWithError` instance."""
        key: str = "error"
        message: Any = "unknown error"

        if isinstance(payload, dict):
            key = "error" if "error" in payload else set(payload.keys()).pop()
            message = payload[key]
            if isinstance(message, list):
                message = message.pop()

        super().__init__(f"Paperless: {key} - {message}")


# Models


class AsnRequestError(PaperlessError):
    """Raise when getting an error during requesting the next asn."""


class DraftFieldRequiredError(PaperlessError):
    """Raise when trying to save models with missing required fields."""


class DraftNotSupportedError(PaperlessError):
    """Raise when trying to draft unsupported models."""


class PrimaryKeyRequiredError(PaperlessError):
    """Raise when trying to access model data without supplying a pk."""


# Tasks


class TaskNotFoundError(PaperlessError):
    """Raise when trying to access a task by non-existing uuid."""

    def __init__(self, task_id: str) -> None:
        """Initialize a `TaskNotFound` instance."""
        super().__init__(f"Task with UUID {task_id} not found.")



"""Provide `User` and 'Group' related models and helpers."""

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .mixins import helpers

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class Group(PaperlessModel):
    """Represent a Paperless `Group`."""

    _api_path = API_PATH["groups_single"]

    id: int
    name: str | None = None
    permissions: list[str] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Group` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class User(PaperlessModel):
    """Represent a Paperless `User`."""

    _api_path = API_PATH["users_single"]

    id: int
    username: str | None = None
    # exclude that from the dataclass
    # password: str | None = None  # noqa: ERA001
    email: str | None = None
    first_name: str | None = None
    last_name: str | None = None
    date_joined: datetime.datetime | None = None
    is_staff: bool | None = None
    is_active: bool | None = None
    is_superuser: bool | None = None
    groups: list[int] | None = None
    user_permissions: list[str] | None = None
    inherited_permissions: list[str] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `User` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class GroupHelper(
    HelperBase[Group],
    helpers.CallableMixin[Group],
    helpers.IterableMixin[Group],
):
    """Represent a factory for Paperless `Group` models."""

    _api_path = API_PATH["groups"]
    _resource = PaperlessResource.GROUPS

    _resource_cls = Group


class UserHelper(
    HelperBase[User],
    helpers.CallableMixin[User],
    helpers.IterableMixin[User],
):
    """Represent a factory for Paperless `User` models."""

    _api_path = API_PATH["users"]
    _resource = PaperlessResource.USERS

    _resource_cls = User



"""Provide the `Paginated` class."""

import math
from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import Any, Generic

from pypaperless.const import API_PATH
from pypaperless.models.base import ResourceT

from .base import PaperlessModel


@dataclass(init=False)
class Page(
    PaperlessModel,
    Generic[ResourceT],
):
    """Represent a Paperless DRF `Paginated`."""

    _api_path = API_PATH["index"]
    _resource_cls: type[ResourceT]

    # our fields
    current_page: int
    page_size: int

    # DRF fields
    count: int
    next: str | None = None
    previous: str | None = None
    all: list[int] = field(default_factory=list)
    results: list[dict[str, Any]] = field(default_factory=list)

    def __iter__(self) -> Iterator[ResourceT]:
        """Return iter of `.items`."""
        return iter(self.items)

    @property
    def current_count(self) -> int:
        """Return the item count of the current page."""
        return len(self.results)

    @property
    def has_next_page(self) -> bool:
        """Return whether there is a next page or not."""
        return self.next_page is not None

    @property
    def has_previous_page(self) -> bool:
        """Return whether there is a previous page or not."""
        return self.previous_page is not None

    @property
    def items(self) -> list[ResourceT]:
        """Return the results list field with mapped PyPaperless `models`.

        Example:
        -------
        ```python
        async for page in paperless.documents.pages():
            assert isinstance(page.results.pop(), Document) # fails, it is a dict
            assert isinstance(page.items.pop(), Document) # ok
        ```

        """

        def mapper(data: dict[str, Any]) -> ResourceT:
            return self._resource_cls.create_with_data(self._api, data, fetched=True)

        return list(map(mapper, self._data["results"]))

    @property
    def is_last_page(self) -> bool:
        """Return whether we are on the last page or not."""
        return not self.has_next_page

    @property
    def last_page(self) -> int:
        """Return the last page number."""
        return math.ceil(self.count / self.page_size)

    @property
    def next_page(self) -> int | None:
        """Return the next page number if a next page exists."""
        if self.next is None:
            return None
        return self.current_page + 1

    @property
    def previous_page(self) -> int | None:
        """Return the previous page number if a previous page exists."""
        if self.previous is None:
            return None
        return self.current_page - 1



"""PyPaperless helpers."""

# pylint: disable=unused-import

from .models.base import HelperBase  # noqa: F401
from .models.classifiers import (  # noqa: F401
    CorrespondentHelper,
    DocumentTypeHelper,
    StoragePathHelper,
    TagHelper,
)
from .models.config import ConfigHelper  # noqa: F401
from .models.custom_fields import CustomFieldHelper  # noqa: F401
from .models.documents import DocumentHelper, DocumentMetaHelper, DocumentNoteHelper  # noqa: F401
from .models.mails import MailAccountHelper, MailRuleHelper  # noqa: F401
from .models.permissions import GroupHelper, UserHelper  # noqa: F401
from .models.saved_views import SavedViewHelper  # noqa: F401
from .models.share_links import ShareLinkHelper  # noqa: F401
from .models.status import StatusHelper  # noqa: F401
from .models.tasks import TaskHelper  # noqa: F401
from .models.workflows import WorkflowHelper  # noqa: F401



"""PyPaperless constants."""

from __future__ import annotations

from enum import StrEnum

CONFIG = "config"
CONSUMPTION_TEMPLATES = "consumption_templates"
CORRESPONDENTS = "correspondents"
CUSTOM_FIELDS = "custom_fields"
DOCUMENTS = "documents"
DOCUMENT_TYPES = "document_types"
GROUPS = "groups"
LOGS = "logs"
MAIL_ACCOUNTS = "mail_accounts"
MAIL_RULES = "mail_rules"
SAVED_VIEWS = "saved_views"
SHARE_LINKS = "share_links"
STATUS = "status"
STORAGE_PATHS = "storage_paths"
TAGS = "tags"
TASKS = "tasks"
USERS = "users"
WORKFLOW_ACTIONS = "workflow_actions"
WORKFLOWS = "workflows"
WORKFLOW_TRIGGERS = "workflow_triggers"
UNKNOWN = "unknown"

API_PATH = {
    "index": "/api/",
    "token": "/api/token/",
    f"{CONFIG}": f"/api/{CONFIG}/",
    f"{CONFIG}_single": f"/api/{CONFIG}/{{pk}}/",
    f"{CORRESPONDENTS}": f"/api/{CORRESPONDENTS}/",
    f"{CORRESPONDENTS}_single": f"/api/{CORRESPONDENTS}/{{pk}}/",
    f"{CUSTOM_FIELDS}": f"/api/{CUSTOM_FIELDS}/",
    f"{CUSTOM_FIELDS}_single": f"/api/{CUSTOM_FIELDS}/{{pk}}/",
    f"{DOCUMENTS}": f"/api/{DOCUMENTS}/",
    f"{DOCUMENTS}_download": f"/api/{DOCUMENTS}/{{pk}}/download/",
    f"{DOCUMENTS}_meta": f"/api/{DOCUMENTS}/{{pk}}/metadata/",
    f"{DOCUMENTS}_next_asn": f"/api/{DOCUMENTS}/next_asn/",
    f"{DOCUMENTS}_notes": f"/api/{DOCUMENTS}/{{pk}}/notes/",
    f"{DOCUMENTS}_preview": f"/api/{DOCUMENTS}/{{pk}}/preview/",
    f"{DOCUMENTS}_thumbnail": f"/api/{DOCUMENTS}/{{pk}}/thumb/",
    f"{DOCUMENTS}_post": f"/api/{DOCUMENTS}/post_document/",
    f"{DOCUMENTS}_single": f"/api/{DOCUMENTS}/{{pk}}/",
    f"{DOCUMENTS}_suggestions": f"/api/{DOCUMENTS}/{{pk}}/suggestions/",
    f"{DOCUMENT_TYPES}": f"/api/{DOCUMENT_TYPES}/",
    f"{DOCUMENT_TYPES}_single": f"/api/{DOCUMENT_TYPES}/{{pk}}/",
    f"{GROUPS}": f"/api/{GROUPS}/",
    f"{GROUPS}_single": f"/api/{GROUPS}/{{pk}}/",
    f"{MAIL_ACCOUNTS}": f"/api/{MAIL_ACCOUNTS}/",
    f"{MAIL_ACCOUNTS}_single": f"/api/{MAIL_ACCOUNTS}/{{pk}}/",
    f"{MAIL_RULES}": f"/api/{MAIL_RULES}/",
    f"{MAIL_RULES}_single": f"/api/{MAIL_RULES}/{{pk}}/",
    f"{SAVED_VIEWS}": f"/api/{SAVED_VIEWS}/",
    f"{SAVED_VIEWS}_single": f"/api/{SAVED_VIEWS}/{{pk}}/",
    f"{SHARE_LINKS}": f"/api/{SHARE_LINKS}/",
    f"{SHARE_LINKS}_single": f"/api/{SHARE_LINKS}/{{pk}}/",
    f"{STATUS}": f"/api/{STATUS}/",
    f"{STORAGE_PATHS}": f"/api/{STORAGE_PATHS}/",
    f"{STORAGE_PATHS}_single": f"/api/{STORAGE_PATHS}/{{pk}}/",
    f"{TAGS}": f"/api/{TAGS}/",
    f"{TAGS}_single": f"/api/{TAGS}/{{pk}}/",
    f"{TASKS}": f"/api/{TASKS}/",
    f"{TASKS}_single": f"/api/{TASKS}/{{pk}}/",
    f"{USERS}": f"/api/{USERS}/",
    f"{USERS}_single": f"/api/{USERS}/{{pk}}/",
    f"{WORKFLOWS}": f"/api/{WORKFLOWS}/",
    f"{WORKFLOWS}_single": f"/api/{WORKFLOWS}/{{pk}}/",
    f"{WORKFLOW_ACTIONS}": f"/api/{WORKFLOW_ACTIONS}/",
    f"{WORKFLOW_ACTIONS}_single": f"/api/{WORKFLOW_ACTIONS}/{{pk}}/",
    f"{WORKFLOW_TRIGGERS}": f"/api/{WORKFLOW_TRIGGERS}/",
    f"{WORKFLOW_TRIGGERS}_single": f"/api/{WORKFLOW_TRIGGERS}/{{pk}}/",
}


class PaperlessResource(StrEnum):
    """Represent paths of api endpoints."""

    CONFIG = CONFIG
    CONSUMPTION_TEMPLATES = CONSUMPTION_TEMPLATES
    CORRESPONDENTS = CORRESPONDENTS
    CUSTOM_FIELDS = CUSTOM_FIELDS
    DOCUMENTS = DOCUMENTS
    DOCUMENT_TYPES = DOCUMENT_TYPES
    GROUPS = GROUPS
    LOGS = LOGS
    MAIL_ACCOUNTS = MAIL_ACCOUNTS
    MAIL_RULES = MAIL_RULES
    SAVED_VIEWS = SAVED_VIEWS
    SHARE_LINKS = SHARE_LINKS
    STATUS = STATUS
    STORAGE_PATHS = STORAGE_PATHS
    TAGS = TAGS
    TASKS = TASKS
    USERS = USERS
    WORKFLOWS = WORKFLOWS
    WORKFLOW_ACTIONS = WORKFLOW_ACTIONS
    WORKFLOW_TRIGGERS = WORKFLOW_TRIGGERS
    UNKNOWN = UNKNOWN

    @classmethod
    def _missing_(cls: type[PaperlessResource], *_: object) -> PaperlessResource:
        """Set default member on unknown value."""
        return cls.UNKNOWN



"""Provide `Workflow` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import WorkflowActionType, WorkflowTriggerSourceType, WorkflowTriggerType
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class WorkflowAction(PaperlessModel):
    """Represent a Paperless `WorkflowAction`."""

    _api_path = API_PATH["workflow_actions_single"]

    id: int | None = None
    type: WorkflowActionType | None = None
    assign_title: str | None = None
    assign_tags: list[int] | None = None
    assign_correspondent: int | None = None
    assign_document_type: int | None = None
    assign_storage_path: int | None = None
    assign_view_users: list[int] | None = None
    assign_view_groups: list[int] | None = None
    assign_change_users: list[int] | None = None
    assign_change_groups: list[int] | None = None
    assign_custom_fields: list[int] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Workflow` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class WorkflowTrigger(
    PaperlessModel,
    models.MatchingFieldsMixin,
):
    """Represent a Paperless `WorkflowTrigger`."""

    _api_path = API_PATH["workflow_triggers_single"]

    id: int | None = None
    sources: list[WorkflowTriggerSourceType] | None = None
    type: WorkflowTriggerType | None = None
    filter_path: str | None = None
    filter_filename: str | None = None
    filter_mailrule: int | None = None
    filter_has_tags: list[int] | None = None
    filter_has_correspondent: int | None = None
    filter_has_document_type: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Workflow` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class Workflow(PaperlessModel):
    """Represent a Paperless `Workflow`."""

    _api_path = API_PATH["workflows_single"]

    id: int | None = None
    name: str | None = None
    order: int | None = None
    enabled: bool | None = None
    actions: list[WorkflowAction] | None = None
    triggers: list[WorkflowTrigger] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Workflow` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class WorkflowActionHelper(
    HelperBase[WorkflowAction],
    helpers.CallableMixin[WorkflowAction],
    helpers.IterableMixin[WorkflowAction],
):
    """Represent a factory for Paperless `WorkflowAction` models."""

    _api_path = API_PATH["workflow_actions"]
    _resource = PaperlessResource.WORKFLOW_ACTIONS

    _resource_cls = WorkflowAction


class WorkflowTriggerHelper(
    HelperBase[WorkflowTrigger],
    helpers.CallableMixin[WorkflowTrigger],
    helpers.IterableMixin[WorkflowTrigger],
):
    """Represent a factory for Paperless `WorkflowTrigger` models."""

    _api_path = API_PATH["workflow_triggers"]
    _resource = PaperlessResource.WORKFLOW_TRIGGERS

    _resource_cls = WorkflowTrigger


class WorkflowHelper(
    HelperBase[Workflow],
    helpers.CallableMixin[Workflow],
    helpers.IterableMixin[Workflow],
):
    """Represent a factory for Paperless `Workflow` models."""

    _api_path = API_PATH["workflows"]
    _resource = PaperlessResource.WORKFLOWS

    _resource_cls = Workflow

    def __init__(self, api: "Paperless") -> None:
        """Initialize a `WorkflowHelper` instance."""
        super().__init__(api)

        self._actions = WorkflowActionHelper(api)
        self._triggers = WorkflowTriggerHelper(api)

    @property
    def actions(self) -> WorkflowActionHelper:
        """Return the attached `WorkflowActionHelper` instance.

        Example:
        -------
        ```python
        wf_action = await paperless.workflows.actions(5)
        ```

        """
        return self._actions

    @property
    def triggers(self) -> WorkflowTriggerHelper:
        """Return the attached `WorkflowTriggerHelper` instance.

        Example:
        -------
        ```python
        wf_trigger = await paperless.workflows.triggers(23)
        ```

        """
        return self._triggers



"""Provide `SavedView` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import SavedViewFilterRuleType
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class SavedView(
    PaperlessModel,
    models.SecurableMixin,
):
    """Represent a Paperless `SavedView`."""

    _api_path = API_PATH["saved_views_single"]

    id: int | None = None
    name: str | None = None
    show_on_dashboard: bool | None = None
    show_in_sidebar: bool | None = None
    sort_field: str | None = None
    sort_reverse: bool | None = None
    filter_rules: list[SavedViewFilterRuleType] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `SavedView` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class SavedViewHelper(
    HelperBase[SavedView],
    helpers.CallableMixin[SavedView],
    helpers.IterableMixin[SavedView],
    helpers.SecurableMixin,
):
    """Represent a factory for Paperless `SavedView` models."""

    _api_path = API_PATH["saved_views"]
    _resource = PaperlessResource.SAVED_VIEWS

    _resource_cls = SavedView



"""CallableMixin for PyPaperless helpers."""

from pypaperless.models.base import HelperProtocol, ResourceT


class CallableMixin(HelperProtocol[ResourceT]):
    """Provide methods for calling a specific resource item."""

    async def __call__(
        self,
        pk: int,
        *,
        lazy: bool = False,
    ) -> ResourceT:
        """Request exactly one resource item.

        Example:
        -------
        ```python
        # request a document
        document = await paperless.documents(42)

        # initialize a model and request it later
        document = await paperless.documents(42, lazy=True)
        ```

        """
        data = {
            "id": pk,
        }
        item = self._resource_cls.create_with_data(self._api, data)

        # set requesting full permissions
        if getattr(self, "_request_full_perms", False):
            item._params.update({"full_perms": "true"})  # noqa: SLF001

        if not lazy:
            await item.load()
        return item



"""Provide `Config` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .mixins import helpers

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class Config(
    PaperlessModel,
):
    """Represent a Paperless `Config`."""

    _api_path = API_PATH["config_single"]

    id: int | None = None
    user_args: str | None = None
    output_type: str | None = None
    pages: int | None = None
    language: str | None = None
    mode: str | None = None
    skip_archive_file: str | None = None
    image_dpi: int | None = None
    unpaper_clean: str | None = None
    deskew: bool | None = None
    rotate_pages: bool | None = None
    rotate_pages_threshold: float | None = None
    max_image_pixels: float | None = None
    color_conversion_strategy: str | None = None
    app_title: str | None = None
    app_logo: str | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Config` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class ConfigHelper(
    HelperBase[Config],
    helpers.CallableMixin[Config],
):
    """Represent a factory for Paperless `Config` models."""

    _api_path = API_PATH["config"]
    _resource = PaperlessResource.CONFIG

    _resource_cls = Config



"""Provide `CustomField` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import CustomFieldType
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class CustomField(
    PaperlessModel,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `CustomField`."""

    _api_path = API_PATH["custom_fields_single"]

    id: int
    name: str | None = None
    data_type: CustomFieldType | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Document` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class CustomFieldDraft(
    PaperlessModel,
    models.CreatableMixin,
):
    """Represent a new Paperless `CustomField`, which is not stored in Paperless."""

    _api_path = API_PATH["custom_fields"]

    _create_required_fields = {"name", "data_type"}

    name: str | None = None
    data_type: CustomFieldType | None = None


class CustomFieldHelper(
    HelperBase[CustomField],
    helpers.CallableMixin[CustomField],
    helpers.DraftableMixin[CustomFieldDraft],
    helpers.IterableMixin[CustomField],
):
    """Represent a factory for Paperless `CustomField` models."""

    _api_path = API_PATH["custom_fields"]
    _resource = PaperlessResource.CUSTOM_FIELDS

    _draft_cls = CustomFieldDraft
    _resource_cls = CustomField



"""Provide `Task` related models and helpers."""

from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.exceptions import TaskNotFoundError

from .base import HelperBase, PaperlessModel
from .common import TaskStatusType

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class Task(
    PaperlessModel,
):
    """Represent a Paperless `Task`."""

    _api_path = API_PATH["tasks_single"]

    id: int | None = None
    task_id: str | None = None
    task_file_name: str | None = None
    date_created: str | None = None
    date_done: str | None = None
    type: str | None = None
    status: TaskStatusType | None = None
    result: str | None = None
    acknowledged: bool | None = None
    related_document: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Task` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class TaskHelper(
    HelperBase[Task],
):
    """Represent a factory for Paperless `Task` models."""

    _api_path = API_PATH["tasks"]
    _resource = PaperlessResource.TASKS

    _resource_cls = Task

    async def __aiter__(self) -> AsyncIterator[Task]:
        """Iterate over task items.

        Example:
        -------
        ```python
        async for task in paperless.tasks:
            # do something
        ```

        """
        res = await self._api.request_json("get", self._api_path)
        for data in res:
            yield self._resource_cls.create_with_data(self._api, data, fetched=True)

    async def __call__(self, task_id: int | str) -> Task:
        """Request exactly one task by id.

        If task_id is `str`: interpret it as a task uuid.
        If task_id is `int`: interpret it as a primary key.

        Example:
        -------
        ```python
        task = await paperless.tasks("uuid-string")
        task = await paperless.tasks(1337)
        ```

        """
        if isinstance(task_id, str):
            params = {
                "task_id": task_id,
            }
            res = await self._api.request_json("get", self._api_path, params=params)
            try:
                item = self._resource_cls.create_with_data(self._api, res.pop(), fetched=True)
            except IndexError as exc:
                raise TaskNotFoundError(task_id) from exc
        else:
            data = {
                "id": task_id,
            }
            item = self._resource_cls.create_with_data(self._api, data)
            await item.load()

        return item



"""PyPaperless models."""

from .classifiers import (
    Correspondent,
    CorrespondentDraft,
    DocumentType,
    DocumentTypeDraft,
    StoragePath,
    StoragePathDraft,
    Tag,
    TagDraft,
)
from .config import Config
from .custom_fields import CustomField, CustomFieldDraft
from .documents import (
    Document,
    DocumentDraft,
    DocumentMeta,
    DocumentNote,
    DocumentNoteDraft,
)
from .mails import MailAccount, MailRule
from .pages import Page
from .permissions import Group, User
from .saved_views import SavedView
from .share_links import ShareLink, ShareLinkDraft
from .status import Status
from .tasks import Task
from .workflows import Workflow, WorkflowAction, WorkflowTrigger

__all__ = (
    "Config",
    "Correspondent",
    "CorrespondentDraft",
    "CustomField",
    "CustomFieldDraft",
    "Document",
    "DocumentDraft",
    "DocumentMeta",
    "DocumentNote",
    "DocumentNoteDraft",
    "DocumentType",
    "DocumentTypeDraft",
    "Group",
    "MailAccount",
    "MailRule",
    "Page",
    "SavedView",
    "ShareLink",
    "ShareLinkDraft",
    "Status",
    "StoragePath",
    "StoragePathDraft",
    "Tag",
    "TagDraft",
    "Task",
    "User",
    "Workflow",
    "WorkflowAction",
    "WorkflowTrigger",
)



"""Provide `Document` related models and helpers."""

import datetime
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.exceptions import AsnRequestError, PrimaryKeyRequiredError
from pypaperless.models.utils import object_to_dict_value

from .base import HelperBase, PaperlessModel
from .common import (
    CustomFieldValueType,
    DocumentMetadataType,
    DocumentSearchHitType,
    RetrieveFileMode,
)
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class Document(
    PaperlessModel,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `Document`."""

    _api_path = API_PATH["documents_single"]

    id: int | None = None
    correspondent: int | None = None
    document_type: int | None = None
    storage_path: int | None = None
    title: str | None = None
    content: str | None = None
    tags: list[int] | None = None
    created: datetime.datetime | None = None
    created_date: datetime.date | None = None
    modified: datetime.datetime | None = None
    added: datetime.datetime | None = None
    archive_serial_number: int | None = None
    original_file_name: str | None = None
    archived_file_name: str | None = None
    is_shared_by_requester: bool | None = None
    custom_fields: list[CustomFieldValueType] | None = None
    __search_hit__: DocumentSearchHitType | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Document` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))
        self.notes = DocumentNoteHelper(api, data.get("id"))

    @property
    def has_search_hit(self) -> bool:
        """Return if the document has a search hit attached."""
        return self.__search_hit__ is not None

    @property
    def search_hit(self) -> DocumentSearchHitType | None:
        """Return the document search hit."""
        return self.__search_hit__

    async def get_download(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.download(cast(int, self.id), original=original)

    async def get_metadata(self) -> "DocumentMeta":
        """Request and return the documents `DocumentMeta` class."""
        return await self._api.documents.metadata(cast(int, self.id))

    async def get_preview(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.preview(cast(int, self.id), original=original)

    async def get_suggestions(self) -> "DocumentSuggestions":
        """Request and return the `DocumentSuggestions` class."""
        return await self._api.documents.suggestions(cast(int, self.id))

    async def get_thumbnail(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.thumbnail(cast(int, self.id), original=original)


@dataclass(init=False)
class DocumentDraft(
    PaperlessModel,
    models.CreatableMixin,
):
    """Represent a new Paperless `Document`, which is not stored in Paperless."""

    _api_path = API_PATH["documents_post"]

    _create_required_fields = {"document"}

    document: bytes | None = None
    filename: str | None = None
    title: str | None = None
    created: datetime.datetime | None = None
    correspondent: int | None = None
    document_type: int | None = None
    storage_path: int | None = None
    tags: int | list[int] | None = None
    archive_serial_number: int | None = None
    custom_fields: list[int] | None = None

    def _serialize(self) -> dict[str, Any]:
        """Serialize."""
        data = {
            "form": {
                field.name: object_to_dict_value(getattr(self, field.name))
                for field in self._get_dataclass_fields()
                if field.name not in {"document", "filename"}
            }
        }
        data["form"].update(
            {
                "document": (
                    (self.document, self.filename) if self.filename is not None else self.document
                )
            }
        )
        return data


@dataclass(init=False)
class DocumentNote(PaperlessModel):
    """Represent a Paperless `DocumentNote`."""

    _api_path = API_PATH["documents_notes"]

    id: int | None = None
    note: str | None = None
    created: datetime.datetime | None = None
    document: int | None = None
    user: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentNote` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("document"))

    async def delete(self) -> bool:
        """Delete a `resource item` from DRF. There is no point of return.

        Return `True` when deletion was successful, `False` otherwise.

        Example:
        -------
        ```python
        # request document notes
        notes = await paperless.documents.notes(42)

        for note in notes:
            if await note.delete():
                print("Successfully deleted the note!")
        ```

        """
        params = {
            "id": self.id,
        }
        async with self._api.request("delete", self._api_path, params=params) as res:
            return res.status == 204


@dataclass(kw_only=True)
class DocumentNoteDraft(
    PaperlessModel,
    models.CreatableMixin,
):
    """Represent a new Paperless `DocumentNote`, which is not stored in Paperless."""

    _api_path = API_PATH["documents_notes"]

    _create_required_fields = {"note", "document"}

    note: str | None = None
    document: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentNote` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("document"))


@dataclass(init=False)
class DocumentMeta(PaperlessModel):
    """Represent a Paperless `Document`s metadata."""

    _api_path = API_PATH["documents_meta"]

    id: int | None = None
    original_checksum: str | None = None
    original_size: int | None = None
    original_mime_type: str | None = None
    media_filename: str | None = None
    has_archive_version: bool | None = None
    original_metadata: list[DocumentMetadataType] | None = None
    archive_checksum: str | None = None
    archive_media_filename: str | None = None
    original_filename: str | None = None
    lang: str | None = None
    archive_size: int | None = None
    archive_metadata: list[DocumentMetadataType] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentMeta` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class DownloadedDocument(PaperlessModel):
    """Represent a Paperless `Document`s downloaded file."""

    _api_path = API_PATH["documents"]

    id: int | None = None
    mode: RetrieveFileMode | None = None
    original: bool | None = None
    content: bytes | None = None
    content_type: str | None = None
    disposition_filename: str | None = None
    disposition_type: str | None = None

    async def load(self) -> None:
        """Get `raw data` from DRF."""
        self._api_path = self._api_path.format(pk=self._data.get("id"))

        params = {
            "original": "true" if self._data.get("original", False) else "false",
        }

        async with self._api.request("get", self._api_path, params=params) as res:
            self._data.update(
                {
                    "content": await res.read(),
                    "content_type": res.content_type,
                }
            )

            if res.content_disposition is not None:
                self._data.update(
                    {
                        "disposition_filename": res.content_disposition.filename,
                        "disposition_type": res.content_disposition.type,
                    }
                )

        self._set_dataclass_fields()
        self._fetched = True


@dataclass(init=False)
class DocumentSuggestions(PaperlessModel):
    """Represent a Paperless `Document` suggestions."""

    _api_path = API_PATH["documents_suggestions"]

    id: int | None = None
    correspondents: list[int] | None = None
    tags: list[int] | None = None
    document_types: list[int] | None = None
    storage_paths: list[int] | None = None
    dates: list[datetime.date] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentSuggestions` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class DocumentSuggestionsHelper(HelperBase[DocumentSuggestions]):
    """Represent a factory for Paperless `DocumentSuggestions` models."""

    _api_path = API_PATH["documents_suggestions"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentSuggestions

    async def __call__(self, pk: int) -> DocumentSuggestions:
        """Request exactly one resource item."""
        data = {
            "id": pk,
        }
        item = self._resource_cls.create_with_data(self._api, data)
        await item.load()

        return item


class DocumentSubHelperBase(
    HelperBase[DownloadedDocument],
):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_suggestions"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DownloadedDocument

    async def __call__(
        self,
        pk: int,
        mode: RetrieveFileMode,
        api_path: str,
        *,
        original: bool,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        data = {
            "id": pk,
            "mode": mode,
            "original": original,
        }
        item = self._resource_cls.create_with_data(self._api, data)
        item._api_path = api_path  # noqa: SLF001
        await item.load()

        return item


class DocumentFileDownloadHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_download"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.DOWNLOAD, self._api_path, original=original
        )


class DocumentFilePreviewHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_preview"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.PREVIEW, self._api_path, original=original
        )


class DocumentFileThumbnailHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_thumbnail"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.THUMBNAIL, self._api_path, original=original
        )


class DocumentMetaHelper(
    HelperBase[DocumentMeta],
    helpers.CallableMixin[DocumentMeta],
):
    """Represent a factory for Paperless `DocumentMeta` models."""

    _api_path = API_PATH["documents_meta"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentMeta


class DocumentNoteHelper(HelperBase[DocumentNote]):
    """Represent a factory for Paperless `DocumentNote` models."""

    _api_path = API_PATH["documents_notes"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentNote

    def __init__(self, api: "Paperless", attached_to: int | None = None) -> None:
        """Initialize a `DocumentHelper` instance."""
        super().__init__(api)

        self._attached_to = attached_to

    async def __call__(
        self,
        pk: int | None = None,
    ) -> list[DocumentNote]:
        """Request and return the documents `DocumentNote` list."""
        doc_pk = self._get_document_pk(pk)
        res = await self._api.request_json("get", self._get_api_path(doc_pk))

        # We have to transform data here slightly.
        # There are two major differences in the data depending on which endpoint is requested.
        # url: documents/{:pk}/ ->
        #       .document -> int
        #       .user -> int
        # url: documents/{:pk}/notes/ ->
        #       .document -> does not exist (so we add it here)
        #       .user -> dict(id=int, username=str, first_name=str, last_name=str)
        return [
            self._resource_cls.create_with_data(
                self._api,
                {
                    **item,
                    "document": doc_pk,
                    "user": item["user"]["id"],
                },
                fetched=True,
            )
            for item in res
        ]

    def _get_document_pk(self, pk: int | None = None) -> int:
        """Return the attached document pk, or the parameter."""
        if not any((self._attached_to, pk)):
            message = f"Accessing {type(self).__name__} data without a primary key."
            raise PrimaryKeyRequiredError(message)
        return cast(int, self._attached_to or pk)

    def _get_api_path(self, pk: int) -> str:
        """Return the formatted api path."""
        return self._api_path.format(pk=pk)

    def draft(self, pk: int | None = None, **kwargs: Any) -> DocumentNoteDraft:
        """Return a fresh and empty `DocumentNoteDraft` instance.

        Example:
        -------
        ```python
        draft = paperless.documents.notes.draft(...)
        # do something
        ```

        """
        kwargs.update({"document": self._get_document_pk(pk)})
        return DocumentNoteDraft.create_with_data(
            self._api,
            data=kwargs,
            fetched=True,
        )


class DocumentHelper(
    HelperBase[Document],
    helpers.SecurableMixin,
    helpers.CallableMixin[Document],
    helpers.DraftableMixin[DocumentDraft],
    helpers.IterableMixin[Document],
):
    """Represent a factory for Paperless `Document` models."""

    _api_path = API_PATH["documents"]
    _resource = PaperlessResource.DOCUMENTS

    _draft_cls = DocumentDraft
    _resource_cls = Document

    def __init__(self, api: "Paperless") -> None:
        """Initialize a `DocumentHelper` instance."""
        super().__init__(api)

        self._download = DocumentFileDownloadHelper(api)
        self._meta = DocumentMetaHelper(api)
        self._notes = DocumentNoteHelper(api)
        self._preview = DocumentFilePreviewHelper(api)
        self._suggestions = DocumentSuggestionsHelper(api)
        self._thumbnail = DocumentFileThumbnailHelper(api)

    @property
    def download(self) -> DocumentFileDownloadHelper:
        """Download the contents of an archived file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.download(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_download()
        ```

        """
        return self._download

    @property
    def metadata(self) -> DocumentMetaHelper:
        """Return the attached `DocumentMetaHelper` instance.

        Example:
        -------
        ```python
        # request metadata of a document directly...
        metadata = await paperless.documents.metadata(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)
        metadata = await doc.get_metadata()
        ```

        """
        return self._meta

    @property
    def notes(self) -> DocumentNoteHelper:
        """Return the attached `DocumentNoteHelper` instance.

        Example:
        -------
        ```python
        # request document notes directly...
        notes = await paperless.documents.notes(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)
        notes = await doc.notes()
        ```

        """
        return self._notes

    @property
    def preview(self) -> DocumentFilePreviewHelper:
        """Preview the contents of an archived file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.preview(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_preview()
        ```

        """
        return self._preview

    @property
    def suggestions(self) -> DocumentSuggestionsHelper:
        """Return the attached `DocumentSuggestionsHelper` instance.

        Example:
        -------
        ```python
        # request document suggestions directly...
        suggestions = await paperless.documents.suggestions(42)

        # ... or by using an already fetched document
        doc = await paperless.suggestions(42)

        suggestions = await doc.get_suggestions()
        ```

        """
        return self._suggestions

    @property
    def thumbnail(self) -> DocumentFileThumbnailHelper:
        """Download the contents of a thumbnail file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.thumbnail(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_thumbnail()
        ```

        """
        return self._thumbnail

    async def get_next_asn(self) -> int:
        """Request the next archive serial number from DRF."""
        async with self._api.request("get", API_PATH["documents_next_asn"]) as res:
            try:
                res.raise_for_status()
                return int(await res.text())
            except Exception as exc:
                raise AsnRequestError from exc

    async def more_like(self, pk: int) -> AsyncGenerator[Document, None]:
        """Lookup documents similar to the given document pk.

        Shortcut function. Same behaviour is possible using `reduce()`.

        Documentation: https://docs.paperless-ngx.com/api/#searching-for-documents
        """
        async with self.reduce(more_like_id=pk):
            async for item in self:
                yield item

    async def search(self, query: str) -> AsyncGenerator[Document, None]:
        """Lookup documents by a search query.

        Shortcut function. Same behaviour is possible using `reduce()`.

        Documentation: https://docs.paperless-ngx.com/usage/#basic-usage_searching
        """
        async with self.reduce(query=query):
            async for item in self:
                yield item



"""Mixins for PyPaperless models."""

from .creatable import CreatableMixin
from .data_fields import MatchingFieldsMixin
from .deletable import DeletableMixin
from .securable import SecurableDraftMixin, SecurableMixin
from .updatable import UpdatableMixin

__all__ = (
    "CreatableMixin",
    "DeletableMixin",
    "MatchingFieldsMixin",
    "UpdatableMixin",
    "SecurableMixin",
    "SecurableDraftMixin",
)



"""Utils for pypaperless models.

Since there are common use-cases in transforming dicts to dataclass et vice-versa,
we borrowed some snippets from aiohue instead of re-inventing the wheel.

pypaperless is meant to be the api library for a Home Assistant integration,
so it should be okay I think.

https://github.com/home-assistant-libs/aiohue/

Thanks for the excellent work, guys!
"""

# mypy: ignore-errors
# pylint: disable=all

import logging
from dataclasses import MISSING, asdict, fields, is_dataclass
from datetime import date, datetime
from enum import Enum
from types import NoneType, UnionType
from typing import TYPE_CHECKING, Any, Union, get_args, get_origin, get_type_hints

import pypaperless.models.base as paperless_base

if TYPE_CHECKING:
    from pypaperless import Paperless


def _str_to_datetime(datetimestr: str) -> datetime:
    """Parse datetime from string."""
    return datetime.fromisoformat(datetimestr.replace("Z", "+00:00"))


def _str_to_date(datestr: str) -> date:
    """Parse date from string."""
    return date.fromisoformat(datestr)


def _dateobj_to_str(value: date | datetime) -> str:
    """Parse string from date objects."""
    return value.isoformat()


def object_to_dict_value(value: Any) -> Any:
    """Convert object values to their correspondending json values."""

    def _clean_value(_value_obj: Any) -> Any:
        if isinstance(_value_obj, dict):
            _value_obj = _clean_dict(_value_obj)
        if isinstance(_value_obj, list):
            _value_obj = _clean_list(_value_obj)
        if isinstance(_value_obj, Enum):
            _value_obj = _value_obj.value
        if isinstance(_value_obj, date | datetime):
            _value_obj = _dateobj_to_str(_value_obj)
        if is_dataclass(_value_obj):
            _value_obj = _clean_dict(asdict(_value_obj))
        return _value_obj

    def _clean_list(_list_obj: list) -> list[Any]:
        final = []
        for list_value in _list_obj:
            final.append(_clean_value(list_value))  # noqa: PERF401
        return final

    def _clean_dict(_dict_obj: dict) -> dict[str, Any]:
        final = {}
        for dict_key, dict_value in _dict_obj.items():
            final[dict_key] = _clean_value(dict_value)
        return final

    return _clean_value(value)


def dict_value_to_object(
    name: str,
    value: Any,
    value_type: Any,
    default: Any = MISSING,
    _api: "Paperless | None" = None,
) -> Any:
    """Try to parse a value from raw (json) data and type annotations.

    Since there are common use-cases in transforming dicts to dataclass et vice-versa,
    we borrowed some snippets from aiohue instead of re-inventing the wheel.

    pypaperless is meant to be the api library for a Home Assistant integration,
    so it should be okay I think.

    https://github.com/home-assistant-libs/aiohue/
    """
    # ruff: noqa: PLR0911, PLR0912
    if isinstance(value_type, str):
        # this shouldn't happen, but just in case
        value_type = get_type_hints(value_type, globals(), locals())

    if isinstance(value, dict) and hasattr(value_type, "from_dict"):
        # always prefer classes that have a from_dict
        return value_type.from_dict(value)

    if value is None and not isinstance(default, type(MISSING)):
        return default
    if value is None and value_type is NoneType:
        return None
    if is_dataclass(value_type) and isinstance(value, dict):
        if _api is not None and paperless_base.PaperlessModel in value_type.__mro__:
            return value_type.create_with_data(api=_api, data=value, fetched=True)
        return value_type(
            **{
                field.name: dict_value_to_object(
                    f"{value_type.__name__}.{field.name}",
                    value.get(field.name),
                    field.type,
                    field.default,
                    _api,
                )
                for field in fields(value_type)
            }
        )
    # get origin value type and inspect one-by-one
    origin: Any = get_origin(value_type)
    if origin in (list, tuple, set) and isinstance(value, list | tuple | set):
        return origin(
            dict_value_to_object(name, subvalue, get_args(value_type)[0], _api=_api)
            for subvalue in value
            if subvalue is not None
        )

    # handle dictionary where we should inspect all values
    if origin is dict:
        subkey_type = get_args(value_type)[0]
        subvalue_type = get_args(value_type)[1]
        return {
            dict_value_to_object(subkey, subkey, subkey_type, _api=_api): dict_value_to_object(
                f"{subkey}.value", subvalue, subvalue_type, _api=_api
            )
            for subkey, subvalue in value.items()
        }
    # handle Union type
    if origin is Union or origin is UnionType:
        # try all possible types
        sub_value_types = get_args(value_type)
        for sub_arg_type in sub_value_types:
            if value is NoneType and sub_arg_type is NoneType:
                return value
            if value == {} and sub_arg_type is NoneType:
                # handle case where optional value is received as empty dict from api
                return None
            # try them all until one succeeds
            try:
                return dict_value_to_object(name, value, sub_arg_type, _api=_api)
            except (KeyError, TypeError, ValueError):
                pass
        # if we get to this point, all possibilities failed
        # find out if we should raise or log this
        err = (
            f"Value {value} of type {type(value)} is invalid for {name}, "
            f"expected value of type {value_type}"
        )
        if NoneType not in sub_value_types:
            # raise exception, we have no idea how to handle this value
            raise TypeError(err)
        # failed to parse the (sub) value but None allowed, log only
        logging.getLogger(__name__).warning(err)
        return None
    if origin is type:
        return get_type_hints(value, globals(), locals())
    # handle Any as value type (which is basically unprocessable)
    if value_type is Any:
        return value
    # raise if value is None and the value is required according to annotations
    if value is None and value_type is not NoneType:
        message = f"`{name}` of type `{value_type}` is required."
        raise KeyError(message)

    try:
        if issubclass(value_type, Enum):
            return value_type(value)
        if issubclass(value_type, datetime):
            return _str_to_datetime(value)
        if issubclass(value_type, date):
            return _str_to_date(value)
    except TypeError:
        # happens if value_type is not a class
        pass

    # common type conversions (e.g. int as string)
    if value_type is float and isinstance(value, int):
        return float(value)
    if value_type is int and isinstance(value, str) and value.isnumeric():
        return int(value)

    # If we reach this point, we could not match the value with the type and we raise
    if not isinstance(value, value_type):
        message = f"Value {value} of type {type(value)} is invalid for {name}, \
            expected value of type {value_type}"
        raise TypeError(message)

    return value



"""DraftableMixin for PyPaperless helpers."""

from typing import Any

from pypaperless.exceptions import DraftNotSupportedError
from pypaperless.models.base import HelperProtocol, ResourceT


class DraftableMixin(HelperProtocol[ResourceT]):
    """Provide the `draft` method for PyPaperless helpers."""

    _draft_cls: type[ResourceT]

    def draft(self, **kwargs: Any) -> ResourceT:
        """Return a fresh and empty `PaperlessModel` instance.

        Example:
        -------
        ```python
        draft = paperless.documents.draft(document=bytes(...), title="New Document")
        # do something
        ```

        """
        if not hasattr(self, "_draft_cls"):
            message = "Helper class has no _draft_cls attribute."
            raise DraftNotSupportedError(message)
        kwargs.update({"id": -1})

        return self._draft_cls.create_with_data(self._api, data=kwargs, fetched=True)



"""UpdatableMixin for PyPaperless models."""

from copy import deepcopy
from typing import Any

from pypaperless.models.base import PaperlessModelProtocol
from pypaperless.models.utils import object_to_dict_value

from .securable import SecurableMixin


class UpdatableMixin(PaperlessModelProtocol):
    """Provide the `update` method for PyPaperless models."""

    _data: dict[str, Any]

    async def update(self, *, only_changed: bool = True) -> bool:
        """Send actually changed `model data` to DRF.

        Return `True` when any attribute was updated, `False` otherwise.

        Example:
        -------
        ```python
        # request a document
        document = await paperless.documents(42)

        document.title = "New Title"
        if await document.update():
            print("Successfully updated a field!")
        ```

        """
        updated = False

        if only_changed:
            updated = await self._patch_fields()
        else:
            updated = await self._put_fields()

        self._set_dataclass_fields()
        return updated

    def _check_permissions_field(self, data: dict) -> None:
        """Check."""
        if SecurableMixin not in type(self).__bases__:
            return
        if not self.has_permissions:  # type: ignore[attr-defined]
            return
        if "permissions" in data:
            data["set_permissions"] = deepcopy(data["permissions"])
            del data["permissions"]

    async def _patch_fields(self) -> bool:
        """Use the http `PATCH` method for updating only changed fields."""
        changed = {}
        for field in self._get_dataclass_fields():
            new_value = object_to_dict_value(getattr(self, field.name))

            if field.name in self._data and new_value != self._data[field.name]:
                changed[field.name] = new_value

        if len(changed) == 0:
            return False

        self._check_permissions_field(changed)

        self._data = await self._api.request_json(
            "patch",
            self._api_path,
            json=changed,
            params=self._params,
        )
        return True

    async def _put_fields(self) -> bool:
        """Use the http `PUT` method to replace all fields."""
        data = {
            field.name: object_to_dict_value(getattr(self, field.name))
            for field in self._get_dataclass_fields()
        }

        self._check_permissions_field(data)

        self._data = await self._api.request_json(
            "put",
            self._api_path,
            json=data,
            params=self._params,
        )
        return True



"""IterableMixin for PyPaperless helpers."""

from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Self

from pypaperless.models.base import HelperProtocol, ResourceT
from pypaperless.models.generators import PageGenerator

if TYPE_CHECKING:
    from pypaperless.models import Page


class IterableMixin(HelperProtocol[ResourceT]):
    """Provide methods for iterating over resource items."""

    _aiter_filters: dict[str, str | int] | None

    async def __aiter__(self) -> AsyncIterator[ResourceT]:
        """Iterate over resource items.

        Example:
        -------
        ```python
        async for item in paperless.documents:
            # do something
        ```

        """
        async for page in self.pages():
            for item in page:
                yield item

    @asynccontextmanager
    async def reduce(
        self: Self,
        **kwargs: str | int,
    ) -> AsyncGenerator[Self, None]:
        """Provide context for iterating over resource items with query parameters.

        `kwargs`: Insert any Paperless api supported filter keywords here.
        You can provide `page` and `page_size` parameters, as well.

        Example:
        -------
        ```python
        filters = {
            "page_size": 1337,
            "title__icontains": "2023",
        }

        async with paperless.documents.reduce(**filters):
            # iterate over resource items ...
            async for item in paperless.documents:
                ...

            # ... or iterate pages as-is
            async for page in paperless.documents.pages():
                ...
        ```

        """
        self._aiter_filters = kwargs
        yield self
        self._aiter_filters = None

    async def all(self) -> list[int]:
        """Return a list of all resource item primary keys.

        When used within a `reduce` context, returns a list of filtered primary keys.
        """
        page = await anext(self.pages(page=1))
        return page.all

    def pages(
        self,
        page: int = 1,
        page_size: int = 150,
    ) -> "AsyncIterator[Page[ResourceT]]":
        """Iterate over resource pages.

        `page`: A page number to start with.
        `page_size`: The page size for each requested batch.

        Example:
        -------
        ```python
        async for item in paperless.documents.pages():
            # do something
        ```

        """
        params: dict[str, Any] = getattr(self, "_aiter_filters", None) or {}

        for param, value in params.items():
            if param.endswith("__in"):
                try:
                    value.extend([])  # throw AttributeError if not a list
                    params[param] = ",".join(map(str, value))
                except AttributeError:
                    # value is not a list, don't modify
                    continue

        params.setdefault("page", page)
        params.setdefault("page_size", page_size)

        # set requesting full permissions
        if getattr(self, "_request_full_perms", False):
            params.update({"full_perms": "true"})

        return PageGenerator(self._api, self._api_path, self._resource_cls, params=params)



"""Provide `Status` related models and helpers."""

from dataclasses import dataclass
from typing import cast

from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.models.common import (
    StatusDatabaseType,
    StatusStorageType,
    StatusTasksType,
    StatusType,
)

from .base import HelperBase, PaperlessModel


@dataclass(init=False)
class Status(PaperlessModel):
    """Represent a Paperless `Status`."""

    _api_path = API_PATH["status"]

    pngx_version: str | None = None
    server_os: str | None = None
    install_type: str | None = None
    storage: StatusStorageType | None = None
    database: StatusDatabaseType | None = None
    tasks: StatusTasksType | None = None

    @property
    def has_errors(self) -> bool:
        """Return whether any status flag is `ERROR`."""
        statuses: list[StatusType] = [
            self.database.status if self.database and self.database.status else StatusType.OK,
            *[
                cast(StatusType, getattr(self.tasks, status, StatusType.OK))
                for status in (
                    "redis_status",
                    "celery_status",
                    "classifier_status",
                )
                if self.tasks
            ],
        ]

        return any(st == StatusType.ERROR for st in statuses)


class StatusHelper(HelperBase[Status]):
    """Represent a factory for the Paperless `Status` model."""

    _api_path = API_PATH["status"]
    _resource = PaperlessResource.STATUS
    _resource_public = False

    _resource_cls = Status

    async def __call__(self) -> Status:
        """Request the `Status` model data."""
        res = await self._api.request_json("get", self._api_path)
        return self._resource_cls.create_with_data(self._api, res, fetched=True)



"""SecurableMixin for PyPaperless models."""

from dataclasses import dataclass

from pypaperless.models.common import PermissionTableType


@dataclass(kw_only=True)
class SecurableMixin:
    """Provide permission fields for PyPaperless models."""

    owner: int | None = None
    user_can_change: bool | None = None
    permissions: PermissionTableType | None = None

    @property
    def has_permissions(self) -> bool:
        """Return if the model data includes the permission field."""
        return self.permissions is not None


@dataclass(kw_only=True)
class SecurableDraftMixin:
    """Provide permission fields for PyPaperless draft models."""

    owner: int | None = None
    set_permissions: PermissionTableType | None = None
